class SignatureError(Exception):
    ...

class SignatureMismatchError(SignatureError):
    ...

class SignatureNotFoundError(SignatureError):
    ...

class UnsupportedSignatureError(SignatureError):
    ...

class IncorrectArgumentsOrderError(ValueError):
    ...
